package java.util.concurrent;

import sun.misc.Contended;
import sun.misc.Unsafe;

public class Exchanger<V>
{
    private static final int ASHIFT = 7;
    private static final int MMASK = 255;
    private static final int SEQ = 256;
    private static final int NCPU;
    static final int FULL;
    private static final int SPINS = 1024;
    private static final Object NULL_ITEM;
    private static final Object TIMED_OUT;
    private final Exchanger.Participant participant;
    private volatile Exchanger.Node[] arena;
    private volatile Exchanger.Node slot;
    private volatile int bound;
    private static final Unsafe U;
    private static final long BOUND;
    private static final long SLOT;
    private static final long MATCH;
    private static final long BLOCKER;
    private static final int ABASE;
    
    private final Object arenaExchange(Object o, boolean b, long time) {
        Exchanger.Node[] arena = this.arena;
        Exchanger.Node o2 = (Exchanger.Node)this.participant.get();
        int index = o2.index;
        while (true) {
            long n;
            Exchanger.Node expected = (Exchanger.Node)Exchanger.U.getObjectVolatile(arena, n = (long)((index << 7) + Exchanger.ABASE));
            if (expected != null && Exchanger.U.compareAndSwapObject(arena, n, expected, null)) {
                Object item = expected.item;
                expected.match = o;
                Thread parked = expected.parked;
                if (parked != null) {
                    Exchanger.U.unpark(parked);
                    return item;
                }
                return item;
            }
            else {
                int bound;
                int n2;
                if (index <= (n2 = ((bound = this.bound) & 0xFF)) && expected == null) {
                    o2.item = o;
                    if (Exchanger.U.compareAndSwapObject(arena, n, null, o2)) {
                        long n3 = (b && n2 == 0) ? (System.nanoTime() + time) : 0L;
                        Thread currentThread = Thread.currentThread();
                        int hash = o2.hash;
                        int n4 = 1024;
                        while (true) {
                            Object match = o2.match;
                            if (match != null) {
                                Exchanger.U.putOrderedObject(o2, Exchanger.MATCH, null);
                                o2.item = null;
                                o2.hash = hash;
                                return match;
                            }
                            if (n4 > 0) {
                                int n5 = hash ^ hash << 1;
                                int n6 = n5 ^ n5 >>> 3;
                                hash = (n6 ^ n6 << 10);
                                if (hash == 0) {
                                    hash = (0x400 | (int)currentThread.getId());
                                }
                                else if (hash < 0 && (--n4 & 0x1FF) == 0x0) {
                                    Thread.yield();
                                }
                                else {
                                    continue;
                                }
                            }
                            else if (Exchanger.U.getObjectVolatile(arena, n) != o2) {
                                n4 = 1024;
                            }
                            else if (!currentThread.isInterrupted() && n2 == 0 && (!b || (time = n3 - System.nanoTime()) > 0L)) {
                                Exchanger.U.putObject(currentThread, Exchanger.BLOCKER, this);
                                o2.parked = currentThread;
                                if (Exchanger.U.getObjectVolatile(arena, n) == o2) {
                                    Exchanger.U.park(false, time);
                                }
                                o2.parked = null;
                                Exchanger.U.putObject(currentThread, Exchanger.BLOCKER, null);
                            }
                            else if (Exchanger.U.getObjectVolatile(arena, n) == o2 && Exchanger.U.compareAndSwapObject(arena, n, o2, null)) {
                                if (n2 != 0) {
                                    Exchanger.U.compareAndSwapInt(this, Exchanger.BOUND, bound, bound + 256 - 1);
                                }
                                o2.item = null;
                                o2.hash = hash;
                                int index2 = o2.index >>> 1;
                                o2.index = index2;
                                index = index2;
                                if (Thread.interrupted()) {
                                    return null;
                                }
                                if (b && n2 == 0 && time <= 0L) {
                                    return Exchanger.TIMED_OUT;
                                }
                                break;
                            }
                            else {
                                continue;
                            }
                        }
                    }
                    else {
                        o2.item = null;
                    }
                }
                else {
                    if (o2.bound != bound) {
                        o2.bound = bound;
                        o2.collides = 0;
                        index = ((index != n2 || n2 == 0) ? n2 : (n2 - 1));
                    }
                    else {
                        int collides;
                        if ((collides = o2.collides) < n2 || n2 == Exchanger.FULL || !Exchanger.U.compareAndSwapInt(this, Exchanger.BOUND, bound, bound + 256 + 1)) {
                            o2.collides = collides + 1;
                            index = ((index == 0) ? n2 : (index - 1));
                        }
                        else {
                            index = n2 + 1;
                        }
                    }
                    o2.index = index;
                }
            }
        }
    }
    
    private final Object slotExchange(Object o, boolean b, long time) {
        Exchanger.Node o2 = (Exchanger.Node)this.participant.get();
        Thread currentThread = Thread.currentThread();
        if (currentThread.isInterrupted()) {
            return null;
        }
        while (true) {
            Exchanger.Node slot;
            if ((slot = this.slot) != null) {
                if (Exchanger.U.compareAndSwapObject(this, Exchanger.SLOT, slot, null)) {
                    Object item = slot.item;
                    slot.match = o;
                    Thread parked = slot.parked;
                    if (parked != null) {
                        Exchanger.U.unpark(parked);
                        return item;
                    }
                    return item;
                }
                else {
                    if (Exchanger.NCPU <= 1 || this.bound != 0 || !Exchanger.U.compareAndSwapInt(this, Exchanger.BOUND, 0, 256)) {
                        continue;
                    }
                    this.arena = new Exchanger.Node[Exchanger.FULL + 2 << 7];
                }
            }
            else {
                if (this.arena != null) {
                    return null;
                }
                o2.item = o;
                if (Exchanger.U.compareAndSwapObject(this, Exchanger.SLOT, null, o2)) {
                    int hash = o2.hash;
                    long n = b ? (System.nanoTime() + time) : 0L;
                    int n2 = (Exchanger.NCPU > 1) ? 1024 : 1;
                    Object match;
                    while ((match = o2.match) == null) {
                        if (n2 > 0) {
                            int n3 = hash ^ hash << 1;
                            int n4 = n3 ^ n3 >>> 3;
                            hash = (n4 ^ n4 << 10);
                            if (hash == 0) {
                                hash = (0x400 | (int)currentThread.getId());
                            }
                            else {
                                if (hash >= 0 || (--n2 & 0x1FF) != 0x0) {
                                    continue;
                                }
                                Thread.yield();
                            }
                        }
                        else if (this.slot != o2) {
                            n2 = 1024;
                        }
                        else if (!currentThread.isInterrupted() && this.arena == null && (!b || (time = n - System.nanoTime()) > 0L)) {
                            Exchanger.U.putObject(currentThread, Exchanger.BLOCKER, this);
                            o2.parked = currentThread;
                            if (this.slot == o2) {
                                Exchanger.U.park(false, time);
                            }
                            o2.parked = null;
                            Exchanger.U.putObject(currentThread, Exchanger.BLOCKER, null);
                        }
                        else {
                            if (!Exchanger.U.compareAndSwapObject(this, Exchanger.SLOT, o2, null)) {
                                continue;
                            }
                            match = ((b && time <= 0L && !currentThread.isInterrupted()) ? Exchanger.TIMED_OUT : null);
                            break;
                        }
                    }
                    Exchanger.U.putOrderedObject(o2, Exchanger.MATCH, null);
                    o2.item = null;
                    o2.hash = hash;
                    return match;
                }
                o2.item = null;
            }
        }
    }
    
    public Exchanger() {
        this.participant = new Exchanger.Participant();
    }
    
    public V exchange(V v) throws InterruptedException {
        v = ((v == null) ? Exchanger.NULL_ITEM : v);
        Object o;
        if ((this.arena != null || (o = this.slotExchange(v, false, 0L)) == null) && (Thread.interrupted() || (o = this.arenaExchange(v, false, 0L)) == null)) {
            throw new InterruptedException();
        }
        return (o == Exchanger.NULL_ITEM) ? null : o;
    }
    
    public V exchange(V v, long n, TimeUnit timeUnit) throws InterruptedException, TimeoutException {
        v = ((v == null) ? Exchanger.NULL_ITEM : v);
        long nanos = timeUnit.toNanos(n);
        Object o;
        if ((this.arena != null || (o = this.slotExchange(v, true, nanos)) == null) && (Thread.interrupted() || (o = this.arenaExchange(v, true, nanos)) == null)) {
            throw new InterruptedException();
        }
        if (o == Exchanger.TIMED_OUT) {
            throw new TimeoutException();
        }
        return (o == Exchanger.NULL_ITEM) ? null : o;
    }
    
    static {
        NCPU = Runtime.getRuntime().availableProcessors();
        FULL = ((Exchanger.NCPU >= 510) ? 255 : (Exchanger.NCPU >>> 1));
        NULL_ITEM = new Object();
        TIMED_OUT = new Object();
        int arrayIndexScale;
        try {
            U = Unsafe.getUnsafe();
            Class<Object> clazz = Exchanger.class;
            Class<Object> clazz2 = Exchanger.Node.class;
            Class<Object> clazz3 = Exchanger.Node[].class;
            Class<Object> clazz4 = Thread.class;
            BOUND = Exchanger.U.objectFieldOffset(clazz.getDeclaredField("bound"));
            SLOT = Exchanger.U.objectFieldOffset(clazz.getDeclaredField("slot"));
            MATCH = Exchanger.U.objectFieldOffset(clazz2.getDeclaredField("match"));
            BLOCKER = Exchanger.U.objectFieldOffset(clazz4.getDeclaredField("parkBlocker"));
            arrayIndexScale = Exchanger.U.arrayIndexScale(clazz3);
            ABASE = Exchanger.U.arrayBaseOffset(clazz3) + 128;
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
        if ((arrayIndexScale & arrayIndexScale - 1) != 0x0 || arrayIndexScale > 128) {
            throw new Error("Unsupported array scale");
        }
        return;
    }
    
    @Contended
    static final class Node
    {
        int index;
        int bound;
        int collides;
        int hash;
        Object item;
        volatile Object match;
        volatile Thread parked;
    }
    
    static final class Participant extends ThreadLocal<Exchanger.Node>
    {
        public Exchanger.Node initialValue() {
            return new Exchanger.Node();
        }
    }
}
